@@ -71,7 +71,7 @@ module MRuby
 
     include Rake::DSL
     include LoadGems
-    attr_accessor :name, :bins, :exts, :file_separator, :build_dir, :gem_clone_dir
+    attr_accessor :name, :bins, :exts, :file_separator, :build_dir, :gem_clone_dir, :defines
     attr_reader :products, :libmruby_core_objs, :libmruby_objs, :gems, :toolchains, :presym, :mrbc_build, :gem_dir_to_repo_url
 
     alias libmruby libmruby_objs
@@ -97,6 +97,7 @@ module MRuby
         @file_separator = '/'
         @build_dir = "#{build_dir}/#{@name}"
         @gem_clone_dir = "#{build_dir}/repos/#{@name}"
+        @defines = []
         @cc = Command::Compiler.new(self, %w(.c), label: "CC")
         @cxx = Command::Compiler.new(self, %w(.cc .cxx .cpp), label: "CXX")
         @objc = Command::Compiler.new(self, %w(.m), label: "OBJC")
@@ -122,6 +123,7 @@ module MRuby
         @enable_test = false
         @enable_lock = true
         @enable_presym = true
+        @enable_benchmark = true
         @mrbcfile_external = false
         @internal = internal
         @toolchains = []
@@ -131,15 +133,18 @@ module MRuby
       end
 
       MRuby::Build.current = current
-      current.instance_eval(&block)
-      if current.libmruby_enabled? && !current.mrbcfile_external?
-        if current.presym_enabled?
-          current.create_mrbc_build if current.host? || current.gems["mruby-bin-mrbc"]
-        elsif current.host?
-          current.build_mrbc_exec
+      begin
+        current.instance_eval(&block)
+      ensure
+        if current.libmruby_enabled? && !current.mrbcfile_external?
+          if current.presym_enabled?
+            current.create_mrbc_build if current.host? || current.gems["mruby-bin-mrbc"]
+          elsif current.host?
+            current.build_mrbc_exec
+          end
         end
+        current.presym = Presym.new(current) if current.presym_enabled?
       end
-      current.presym = Presym.new(current) if current.presym_enabled?
     end
 
     def libmruby_enabled?
@@ -228,6 +233,14 @@ module MRuby
       @cxx_abi_enabled = true
     end
 
+    def benchmark_enabled?
+      @enable_benchmark
+    end
+
+    def disable_benchmark
+      @enable_benchmark = false
+    end
+
     def compile_as_cxx(src, cxx_src = nil, obj = nil, includes = [])
       #
       # If `cxx_src` is specified, this method behaves the same as before as
@@ -248,7 +261,7 @@ module MRuby
           end
         end
       else
-        cxx_src = "#{build_dir}/#{src.relative_path})".ext << "-cxx.cxx"
+        cxx_src = "#{build_dir}/#{src.relative_path.to_s.remove_leading_parents}".ext << "-cxx.cxx"
         obj = cxx_src.ext(@exts.object)
       end
 
@@ -320,12 +333,16 @@ EOS
       return @mrbcfile if @mrbcfile
 
       gem_name = "mruby-bin-mrbc"
-      gem = @gems[gem_name]
-      gem ||= (host = MRuby.targets["host"]) && host.gems[gem_name]
-      unless gem
-        fail "external mrbc or mruby-bin-mrbc gem in current('#{@name}') or 'host' build is required"
+      if (gem = @gems[gem_name])
+        @mrbcfile = exefile("#{gem.build.build_dir}/bin/mrbc")
+      elsif !host? && (host = MRuby.targets["host"])
+        if (gem = host.gems[gem_name])
+          @mrbcfile = exefile("#{gem.build.build_dir}/bin/mrbc")
+        elsif host.mrbcfile_external?
+          @mrbcfile = host.mrbcfile
+        end
       end
-      @mrbcfile = exefile("#{gem.build.build_dir}/bin/mrbc")
+      @mrbcfile || fail("external mrbc or mruby-bin-mrbc gem in current('#{@name}') or 'host' build is required")
     end
 
     def mrbcfile=(path)
@@ -344,14 +361,8 @@ EOS
     end
 
     def define_rules
-      use_mrdb = @gems["mruby-bin-debugger"]
       compilers.each do |compiler|
-        if respond_to?(:enable_gems?) && enable_gems?
-          compiler.defines -= %w(MRB_NO_GEMS)
-        else
-          compiler.defines += %w(MRB_NO_GEMS)
-        end
-        compiler.defines |= %w(MRB_USE_DEBUG_HOOK) if use_mrdb
+        compiler.defines << "MRB_NO_GEMS" unless enable_gems? && libmruby_enabled?
       end
       [@cc, *(@cxx if cxx_exception_enabled?)].each do |compiler|
         compiler.define_rules(@build_dir, MRUBY_ROOT, @exts.object)
@@ -489,7 +500,8 @@ EOS
         v = instance_variable_get(n)
         v = case v
             when nil, true, false, Numeric; v
-            when String, Command; v.clone
+            when String; v.clone
+            when Command; v.clone.tap { |u| u.build = build }
             else Marshal.load(Marshal.dump(v))  # deep clone
             end
         build.instance_variable_set(n, v)
@@ -539,6 +551,23 @@ EOS
       end
     end
 
+    def run_bintest
+      puts ">>> Bintest #{name} <<<"
+      targets = @gems.select { |v| File.directory? "#{v.dir}/bintest" }.map { |v| filename v.dir }
+      targets << filename(".") if File.directory? "./bintest"
+      mrbc = @gems["mruby-bin-mrbc"] ? exefile("#{@build_dir}/bin/mrbc") : mrbcfile
+
+      emulator = @test_runner.command
+      emulator = @test_runner.shellquote(emulator) if emulator
+
+      env = {
+        "BUILD_DIR" => @build_dir,
+        "MRBCFILE" => mrbc,
+        "EMULATOR" => @test_runner.emulator,
+      }
+      sh env, "ruby test/bintest.rb#{verbose_flag} #{targets.join ' '}"
+    end
+
     protected
 
     def create_mrbc_build; end
