@@ -47,7 +47,7 @@ end
 
 assert('Hash#merge!') do
   a = { 'abc_key' => 'abc_value', 'cba_key' => 'cba_value' }
-  b = { 'cba_key' => 'XXX',  'xyz_key' => 'xyz_value' }
+  b = { 'cba_key' => 'XXX', 'xyz_key' => 'xyz_value' }
 
   result_1 = a.merge! b
 
@@ -64,6 +64,9 @@ assert('Hash#merge!') do
   assert_raise(TypeError) do
     { 'abc_key' => 'abc_value' }.merge! "a"
   end
+
+  # multiple arguments
+  assert_equal({a:1,b:2,c:3}, {a:1}.merge!({b:2},{c:3}))
 end
 
 assert('Hash#values_at') do
