@@ -14,7 +14,6 @@
 #include <sys/stat.h>
 
 #include <fcntl.h>
-#include <limits.h>
 
 #include <errno.h>
 #include <stdlib.h>
@@ -89,7 +88,7 @@ flock(int fd, int operation) {
   DWORD flags;
   flags = ((operation & LOCK_NB) ? LOCKFILE_FAIL_IMMEDIATELY : 0)
           | ((operation & LOCK_SH) ? LOCKFILE_EXCLUSIVE_LOCK : 0);
-  memset(&ov, 0, sizeof(ov));
+  ov = (OVERLAPPED){0};
   return LockFileEx(h, flags, 0, 0xffffffff, 0xffffffff, &ov) ? 0 : -1;
 }
 #endif
@@ -117,14 +116,14 @@ static mrb_value
 mrb_file_s_unlink(mrb_state *mrb, mrb_value obj)
 {
   const mrb_value *argv;
-  mrb_value pathv;
   mrb_int argc, i;
   char *path;
 
   mrb_get_args(mrb, "*", &argv, &argc);
   for (i = 0; i < argc; i++) {
     const char *utf8_path;
-    pathv = mrb_ensure_string_type(mrb, argv[i]);
+    mrb_value pathv = argv[i];
+    mrb_ensure_string_type(mrb, pathv);
     utf8_path = RSTRING_CSTR(mrb, pathv);
     path = mrb_locale_from_utf8(utf8_path, -1);
     if (UNLINK(path) < 0) {
@@ -262,7 +261,7 @@ mrb_file_realpath(mrb_state *mrb, mrb_value klass)
   result = mrb_str_new_capa(mrb, PATH_MAX);
   if (realpath(cpath, RSTRING_PTR(result)) == NULL) {
     mrb_locale_free(cpath);
-    mrb_sys_fail(mrb, cpath);
+    mrb_sys_fail(mrb, RSTRING_CSTR(mrb, pathname));
     return result;              /* not reached */
   }
   mrb_locale_free(cpath);
@@ -497,7 +496,7 @@ mrb_file_truncate(mrb_state *mrb, mrb_value self)
   mrb_value lenv = mrb_get_arg1(mrb);
 
   fd = mrb_io_fileno(mrb, self);
-  length = mrb_int(mrb, lenv);
+  length = mrb_as_int(mrb, lenv);
   if (mrb_ftruncate(fd, length) != 0) {
     mrb_raise(mrb, E_IO_ERROR, "ftruncate failed");
   }
@@ -593,9 +592,9 @@ mrb_init_file(mrb_state *mrb)
 {
   struct RClass *io, *file, *cnst;
 
-  io   = mrb_class_get(mrb, "IO");
+  io   = mrb_class_get_id(mrb, MRB_SYM(IO));
   file = mrb_define_class(mrb, "File", io);
-  MRB_SET_INSTANCE_TT(file, MRB_TT_DATA);
+  MRB_SET_INSTANCE_TT(file, MRB_TT_CDATA);
   mrb_define_class_method(mrb, file, "umask",  mrb_file_s_umask, MRB_ARGS_OPT(1));
   mrb_define_class_method(mrb, file, "delete", mrb_file_s_unlink, MRB_ARGS_ANY());
   mrb_define_class_method(mrb, file, "unlink", mrb_file_s_unlink, MRB_ARGS_ANY());
