@@ -5,6 +5,7 @@
 #include "mruby/proc.h"
 #include "mruby/class.h"
 #include "mruby/string.h"
+#include <mruby/internal.h>
 #include "mruby/presym.h"
 
 typedef enum {
@@ -21,6 +22,8 @@ typedef enum {
   NOEX_RESPONDS  = 0x80
 } mrb_method_flag_t;
 
+mrb_value mrb_proc_local_variables(mrb_state *mrb, const struct RProc *proc);
+
 static mrb_value
 mrb_f_nil(mrb_state *mrb, mrb_value cv)
 {
@@ -133,40 +136,7 @@ mrb_obj_ivar_set(mrb_state *mrb, mrb_value self)
 static mrb_value
 mrb_local_variables(mrb_state *mrb, mrb_value self)
 {
-  const struct RProc *proc;
-  const mrb_irep *irep;
-  mrb_value vars;
-  size_t i;
-
-  proc = mrb->c->ci[-1].proc;
-
-  if (proc == NULL || MRB_PROC_CFUNC_P(proc)) {
-    return mrb_ary_new(mrb);
-  }
-  vars = mrb_hash_new(mrb);
-  while (proc) {
-    if (MRB_PROC_CFUNC_P(proc)) break;
-    irep = proc->body.irep;
-    if (irep->lv) {
-      for (i = 0; i + 1 < irep->nlocals; ++i) {
-        if (irep->lv[i]) {
-          mrb_sym sym = irep->lv[i];
-          const char *name = mrb_sym_name(mrb, sym);
-          switch (name[0]) {
-          case '*': case '&':
-            break;
-          default:
-            mrb_hash_set(mrb, vars, mrb_symbol_value(sym), mrb_true_value());
-            break;
-          }
-        }
-      }
-    }
-    if (MRB_PROC_SCOPE_P(proc)) break;
-    proc = proc->upper;
-  }
-
-  return mrb_hash_keys(mrb, vars);
+  return mrb_proc_local_variables(mrb, mrb->c->ci[-1].proc);
 }
 
 KHASH_DECLARE(st, mrb_sym, char, FALSE)
@@ -205,41 +175,37 @@ method_entry_loop(mrb_state *mrb, struct RClass *klass, khash_t(st) *set, khash_
 }
 
 static mrb_value
-mrb_class_instance_method_list(mrb_state *mrb, mrb_bool recur, struct RClass *klass, int obj)
+mrb_class_instance_method_list(mrb_state *mrb, mrb_bool recur, struct RClass *klass)
 {
-  khint_t i;
   mrb_value ary;
-  mrb_bool prepended = FALSE;
   struct RClass *oldklass;
   khash_t(st) *set = kh_init(st, mrb);
-  khash_t(st) *undef = (recur ? kh_init(st, mrb) : NULL);
 
-  if (!recur && (klass->flags & MRB_FL_CLASS_IS_PREPENDED)) {
-    MRB_CLASS_ORIGIN(klass);
-    prepended = TRUE;
+  if (!recur) {
+    if (klass->flags & MRB_FL_CLASS_IS_PREPENDED) {
+      MRB_CLASS_ORIGIN(klass);
+    }
+    method_entry_loop(mrb, klass, set, NULL);
   }
+  else {
+    khash_t(st) *undef = kh_init(st, mrb);
 
-  oldklass = 0;
-  while (klass && (klass != oldklass)) {
-    method_entry_loop(mrb, klass, set, undef);
-    if ((klass->tt == MRB_TT_ICLASS && !prepended) ||
-        (klass->tt == MRB_TT_SCLASS)) {
+    oldklass = NULL;
+    while (klass && (klass != oldklass)) {
+      method_entry_loop(mrb, klass, set, undef);
+      oldklass = klass;
+      klass = klass->super;
     }
-    else {
-      if (!recur) break;
-    }
-    oldklass = klass;
-    klass = klass->super;
+    kh_destroy(st, mrb, undef);
   }
 
   ary = mrb_ary_new_capa(mrb, kh_size(set));
-  for (i=0;i<kh_end(set);i++) {
+  for (khint_t i=0; i<kh_end(set); i++) {
     if (kh_exist(set, i)) {
       mrb_ary_push(mrb, ary, mrb_symbol_value(kh_key(set, i)));
     }
   }
   kh_destroy(st, mrb, set);
-  if (undef) kh_destroy(st, mrb, undef);
 
   return ary;
 }
@@ -247,7 +213,7 @@ mrb_class_instance_method_list(mrb_state *mrb, mrb_bool recur, struct RClass *kl
 static mrb_value
 mrb_obj_methods(mrb_state *mrb, mrb_bool recur, mrb_value obj, mrb_method_flag_t flag)
 {
-  return mrb_class_instance_method_list(mrb, recur, mrb_class(mrb, obj), 0);
+  return mrb_class_instance_method_list(mrb, recur, mrb_class(mrb, obj));
 }
 /* 15.3.1.3.31 */
 /*
@@ -607,7 +573,40 @@ mrb_mod_instance_methods(mrb_state *mrb, mrb_value mod)
   struct RClass *c = mrb_class_ptr(mod);
   mrb_bool recur = TRUE;
   mrb_get_args(mrb, "|b", &recur);
-  return mrb_class_instance_method_list(mrb, recur, c, 0);
+  return mrb_class_instance_method_list(mrb, recur, c);
+}
+
+static int
+undefined_method_i(mrb_state *mrb, mrb_sym mid, mrb_method_t m, void *p)
+{
+  mrb_value ary = *(mrb_value*)p;
+
+  if (MRB_METHOD_UNDEF_P(m)) {
+    mrb_ary_push(mrb, ary, mrb_symbol_value(mid));
+  }
+  return 0;
+}
+
+/*
+ *  call-seq:
+ *     mod.undefined_methods()   -> array
+ *
+ *  Returns an array containing the names of the undefined methods of the module/class.
+ */
+static mrb_value
+mrb_mod_undefined_methods(mrb_state *mrb, mrb_value mod)
+{
+  struct RClass *m = mrb_class_ptr(mod);
+  mrb_get_args(mrb, "");        /* no argument */
+
+  mrb_value ary = mrb_ary_new(mrb);
+
+  if (m->flags & MRB_FL_CLASS_IS_PREPENDED) {
+    MRB_CLASS_ORIGIN(m);
+  }
+  mrb_mt_foreach(mrb, m, undefined_method_i, (void*)&ary);
+
+  return ary;
 }
 
 /* 15.2.2.4.41 */
@@ -666,9 +665,6 @@ mrb_mod_s_nesting(mrb_state *mrb, mrb_value mod)
   return ary;
 }
 
-/* implementation of #send method */
-mrb_value mrb_f_send(mrb_state *mrb, mrb_value self);
-
 void
 mrb_mruby_metaprog_gem_init(mrb_state* mrb)
 {
@@ -698,6 +694,7 @@ mrb_mruby_metaprog_gem_init(mrb_state* mrb)
   mrb_define_method(mrb, mod, "class_variable_set", mrb_mod_cvar_set, MRB_ARGS_REQ(2)); /* 15.2.2.4.18 */
   mrb_define_method(mrb, mod, "included_modules", mrb_mod_included_modules, MRB_ARGS_NONE()); /* 15.2.2.4.30 */
   mrb_define_method(mrb, mod, "instance_methods", mrb_mod_instance_methods, MRB_ARGS_ANY()); /* 15.2.2.4.33 */
+  mrb_define_method(mrb, mod, "undefined_instance_methods", mrb_mod_undefined_methods, MRB_ARGS_NONE());
   mrb_define_method(mrb, mod, "remove_method", mrb_mod_remove_method, MRB_ARGS_ANY()); /* 15.2.2.4.41 */
   mrb_define_method(mrb, mod, "method_removed", mrb_f_nil, MRB_ARGS_REQ(1));
   mrb_define_method(mrb, mod, "constants", mrb_mod_constants, MRB_ARGS_OPT(1)); /* 15.2.2.4.24 */
