@@ -4,12 +4,146 @@
 #include "mruby/variable.h"
 #include "mruby/proc.h"
 #include "mruby/string.h"
+#include "mruby/internal.h"
 #include "mruby/presym.h"
 
+// Defined by mruby-proc-ext on which mruby-method depends
+mrb_value mrb_proc_parameters(mrb_state *mrb, mrb_value proc);
+mrb_value mrb_proc_source_location(mrb_state *mrb, struct RProc *p);
+
+static mrb_value
+args_shift(mrb_state *mrb)
+{
+  mrb_callinfo *ci = mrb->c->ci;
+  mrb_value *argv = ci->stack + 1;
+
+  if (ci->n < 15) {
+    if (ci->n == 0) { goto argerr; }
+    mrb_assert(ci->nk == 0 || ci->nk == 15);
+    mrb_value obj = argv[0];
+    int count = ci->n + (ci->nk == 0 ? 0 : 1) + 1 /* block */ - 1 /* first value */;
+    memmove(argv, argv + 1, count * sizeof(mrb_value));
+    ci->n--;
+    return obj;
+  }
+  else if (RARRAY_LEN(*argv) > 0) {
+    return mrb_ary_shift(mrb, *argv);
+  }
+  else {
+  argerr:
+    mrb_argnum_error(mrb, 0, 1, -1);
+    return mrb_undef_value(); /* not reached */
+  }
+}
+
+static void
+args_unshift(mrb_state *mrb, mrb_value obj)
+{
+  mrb_callinfo *ci = mrb->c->ci;
+  mrb_value *argv = ci->stack + 1;
+
+  if (ci->n < 15) {
+    mrb_assert(ci->nk == 0 || ci->nk == 15);
+    mrb_value args = mrb_ary_new_from_values(mrb, ci->n, argv);
+    if (ci->nk == 0) {
+      mrb_value block = argv[ci->n];
+      argv[0] = args;
+      argv[1] = block;
+    }
+    else {
+      mrb_value keyword = argv[ci->n];
+      mrb_value block = argv[ci->n + 1];
+      argv[0] = args;
+      argv[1] = keyword;
+      argv[2] = block;
+    }
+    ci->n = 15;
+  }
+
+  mrb_ary_unshift(mrb, *argv, obj);
+}
+
+static struct RProc*
+method_missing_prepare(mrb_state *mrb, mrb_sym *mid, mrb_value recv, struct RClass **tc)
+{
+  const mrb_sym id_method_missing = MRB_SYM(method_missing);
+  mrb_callinfo *ci = mrb->c->ci;
+
+  if (*mid == id_method_missing) {
+  method_missing: ;
+    int n = ci->n;
+    mrb_value *argv = ci->stack + 1;
+    mrb_value args = (n == 15) ? argv[0] : mrb_ary_new_from_values(mrb, n, argv);
+    mrb_method_missing(mrb, id_method_missing, recv, args);
+  }
+
+  *tc = mrb_class(mrb, recv);
+  mrb_method_t m = mrb_method_search_vm(mrb, tc, id_method_missing);
+  if (MRB_METHOD_UNDEF_P(m)) {
+    goto method_missing;
+  }
+
+  struct RProc *proc;
+  if (MRB_METHOD_FUNC_P(m)) {
+    proc = mrb_proc_new_cfunc(mrb, MRB_METHOD_FUNC(m));
+    MRB_PROC_SET_TARGET_CLASS(proc, *tc);
+  }
+  else {
+    proc = MRB_METHOD_PROC(m);
+  }
+
+  args_unshift(mrb, mrb_symbol_value(*mid));
+  *mid = id_method_missing;
+
+  return proc;
+}
+
 static struct RObject *
 method_object_alloc(mrb_state *mrb, struct RClass *mclass)
 {
-  return (struct RObject*)mrb_obj_alloc(mrb, MRB_TT_OBJECT, mclass);
+  return MRB_OBJ_ALLOC(mrb, MRB_TT_OBJECT, mclass);
+}
+
+static struct RProc*
+method_extract_proc(mrb_state *mrb, mrb_value self)
+{
+  mrb_value obj = mrb_iv_get(mrb, self, MRB_SYM(_proc));
+  if (mrb_nil_p(obj)) {
+    return NULL;
+  }
+  else {
+    mrb_check_type(mrb, obj, MRB_TT_PROC);
+    return mrb_proc_ptr(obj);
+  }
+}
+
+static mrb_value
+method_extract_receiver(mrb_state *mrb, mrb_value self)
+{
+  return mrb_iv_get(mrb, self, MRB_SYM(_recv));
+}
+
+static mrb_sym
+method_extract_mid(mrb_state *mrb, mrb_value self)
+{
+  mrb_value obj = mrb_iv_get(mrb, self, MRB_SYM(_name));
+  mrb_check_type(mrb, obj, MRB_TT_SYMBOL);
+  return mrb_symbol(obj);
+}
+
+static struct RClass*
+method_extract_owner(mrb_state *mrb, mrb_value self)
+{
+  mrb_value obj = mrb_iv_get(mrb, self, MRB_SYM(_owner));
+  switch (mrb_type(obj)) {
+    case MRB_TT_CLASS:
+    case MRB_TT_MODULE:
+    case MRB_TT_SCLASS:
+      break;
+    default:
+      mrb_raise(mrb, E_TYPE_ERROR, "not class/module as owner of method object");
+  }
+  return mrb_class_ptr(obj);
 }
 
 static void
@@ -53,7 +187,7 @@ method_eql(mrb_state *mrb, mrb_value self)
 {
   mrb_value other = mrb_get_arg1(mrb);
   mrb_value receiver, orig_proc, other_proc;
-  struct RClass *owner, *klass;
+  struct RClass *owner;
   struct RProc *orig_rproc, *other_rproc;
 
   if (!mrb_obj_is_instance_of(mrb, other, mrb_class(mrb, self)))
@@ -62,10 +196,6 @@ method_eql(mrb_state *mrb, mrb_value self)
   if (mrb_class(mrb, self) != mrb_class(mrb, other))
     return mrb_false_value();
 
-  klass = mrb_class_ptr(IV_GET(self, MRB_SYM(_klass)));
-  if (klass != mrb_class_ptr(IV_GET(other, MRB_SYM(_klass))))
-    return mrb_false_value();
-
   owner = mrb_class_ptr(IV_GET(self, MRB_SYM(_owner)));
   if (owner != mrb_class_ptr(IV_GET(other, MRB_SYM(_owner))))
     return mrb_false_value();
@@ -83,10 +213,8 @@ method_eql(mrb_state *mrb, mrb_value self)
       return mrb_false_value();
   }
 
-  if (mrb_nil_p(orig_proc))
-    return mrb_false_value();
-  if (mrb_nil_p(other_proc))
-    return mrb_false_value();
+  if (mrb_nil_p(orig_proc)) return mrb_false_value();
+  if (mrb_nil_p(other_proc)) return mrb_false_value();
 
   orig_rproc = mrb_proc_ptr(orig_proc);
   other_rproc = mrb_proc_ptr(other_proc);
@@ -109,61 +237,40 @@ method_eql(mrb_state *mrb, mrb_value self)
 #undef IV_GET
 
 static mrb_value
-mcall(mrb_state *mrb, mrb_value recv, mrb_value proc, mrb_value name, struct RClass *owner,
-      mrb_int argc, const mrb_value *argv, mrb_value block)
+mcall(mrb_state *mrb, mrb_value self, mrb_value recv)
 {
-  mrb_value ret;
-  mrb_sym orig_mid = mrb->c->ci->mid;
+  struct RProc *proc = method_extract_proc(mrb, self);
+  mrb_sym mid = method_extract_mid(mrb, self);
+  struct RClass *tc = method_extract_owner(mrb, self);
 
-  mrb->c->ci->mid = mrb_symbol(name);
-  if (mrb_nil_p(proc)) {
-    mrb_value missing_argv = mrb_ary_new_from_values(mrb, argc, argv);
-    mrb_ary_unshift(mrb, missing_argv, name);
-    ret = mrb_funcall_argv(mrb, recv, MRB_SYM(method_missing), argc + 1, RARRAY_PTR(missing_argv));
-  }
-  else if (!mrb_nil_p(block)) {
-    /*
-      workaround since `mrb_yield_with_class` does not support passing block as parameter
-      need new API that initializes `mrb->c->stack[argc+1]` with block passed by argument
-    */
-    ret = mrb_funcall_with_block(mrb, recv, mrb_symbol(name), argc, argv, block);
+  if (mrb_undef_p(recv)) {
+    recv = method_extract_receiver(mrb, self);
   }
   else {
-    ret = mrb_yield_with_class(mrb, proc, argc, argv, recv, owner);
+    bind_check(mrb, recv, mrb_obj_value(tc));
   }
-  mrb->c->ci->mid = orig_mid;
-  return ret;
+
+  if (!proc) {
+    proc = method_missing_prepare(mrb, &mid, recv, &tc);
+  }
+  mrb->c->ci->mid = mid;
+  mrb->c->ci->u.target_class = tc;
+
+  return mrb_exec_irep(mrb, recv, proc);
 }
 
 static mrb_value
 method_call(mrb_state *mrb, mrb_value self)
 {
-  mrb_value proc = mrb_iv_get(mrb, self, MRB_SYM(_proc));
-  mrb_value name = mrb_iv_get(mrb, self, MRB_SYM(_name));
-  mrb_value recv = mrb_iv_get(mrb, self, MRB_SYM(_recv));
-  struct RClass *owner = mrb_class_ptr(mrb_iv_get(mrb, self, MRB_SYM(_owner)));
-  mrb_int argc;
-  const mrb_value *argv;
-  mrb_value block;
-
-  mrb_get_args(mrb, "*&", &argv, &argc, &block);
-  return mcall(mrb, recv, proc, name, owner, argc, argv, block);
+  return mcall(mrb, self, mrb_undef_value());
 }
 
 static mrb_value
 method_bcall(mrb_state *mrb, mrb_value self)
 {
-  mrb_value proc = mrb_iv_get(mrb, self, MRB_SYM(_proc));
-  mrb_value name = mrb_iv_get(mrb, self, MRB_SYM(_name));
-  mrb_value recv = mrb_iv_get(mrb, self, MRB_SYM(_recv));
-  mrb_value owner = mrb_iv_get(mrb, self, MRB_SYM(_owner));
-  mrb_int argc;
-  const mrb_value *argv;
-  mrb_value block;
-
-  mrb_get_args(mrb, "o*&", &recv, &argv, &argc, &block);
-  bind_check(mrb, recv, owner);
-  return mcall(mrb, recv, proc, name, mrb_class_ptr(owner), argc, argv, block);
+  mrb_value recv = args_shift(mrb);
+  mrb_gc_protect(mrb, recv);
+  return mcall(mrb, self, recv);
 }
 
 static mrb_value
@@ -175,7 +282,7 @@ method_unbind(mrb_state *mrb, mrb_value self)
   mrb_value proc = mrb_iv_get(mrb, self, MRB_SYM(_proc));
   mrb_value klass = mrb_iv_get(mrb, self, MRB_SYM(_klass));
 
-  ume = method_object_alloc(mrb, mrb_class_get(mrb, "UnboundMethod"));
+  ume = method_object_alloc(mrb, mrb_class_get_id(mrb, MRB_SYM(UnboundMethod)));
   mrb_obj_iv_set(mrb, ume, MRB_SYM(_owner), owner);
   mrb_obj_iv_set(mrb, ume, MRB_SYM(_recv), mrb_nil_value());
   mrb_obj_iv_set(mrb, ume, MRB_SYM(_name), name);
@@ -193,7 +300,12 @@ method_search_vm(mrb_state *mrb, struct RClass **cp, mrb_sym mid)
     return NULL;
   if (MRB_METHOD_PROC_P(m))
     return MRB_METHOD_PROC(m);
-  return mrb_proc_new_cfunc(mrb, MRB_METHOD_FUNC(m));
+
+  struct RProc *proc = mrb_proc_new_cfunc(mrb, MRB_METHOD_FUNC(m));
+  if (MRB_METHOD_NOARG_P(m)) {
+    proc->flags |= MRB_PROC_NOARG;
+  }
+  return proc;
 }
 
 static mrb_value
@@ -207,25 +319,25 @@ method_super_method(mrb_state *mrb, mrb_value self)
   struct RProc *proc;
   struct RObject *me;
 
-  switch (mrb_type(klass)) {
-    case MRB_TT_SCLASS:
-      super = mrb_class_ptr(klass)->super->super;
-      break;
-    case MRB_TT_ICLASS:
-      super = mrb_class_ptr(klass)->super;
-      break;
-    default:
-      super = mrb_class_ptr(owner)->super;
-      break;
+  if (mrb_type(owner) == MRB_TT_MODULE) {
+    struct RClass *m = mrb_class_ptr(owner);
+    rklass = mrb_class_ptr(klass)->super;
+    while (rklass && rklass->c != m) {
+      rklass = rklass->super;
+    }
+    if (!rklass) return mrb_nil_value();
+    super = rklass->super;
+  }
+  else {
+    super = mrb_class_ptr(owner)->super;
   }
 
   proc = method_search_vm(mrb, &super, mrb_symbol(name));
-  if (!proc)
-    return mrb_nil_value();
+  if (!proc) return mrb_nil_value();
 
   rklass = super;
-  while (super->tt == MRB_TT_ICLASS)
-    super = super->c;
+  super = mrb_class_real(super);
+  if (!super) return mrb_nil_value();
 
   me = method_object_alloc(mrb, mrb_obj_class(mrb, self));
   mrb_obj_iv_set(mrb, me, MRB_SYM(_owner), mrb_obj_value(super));
@@ -249,28 +361,17 @@ static mrb_value
 method_source_location(mrb_state *mrb, mrb_value self)
 {
   mrb_value proc = mrb_iv_get(mrb, self, MRB_SYM(_proc));
-  struct RProc *rproc;
-  struct RClass *orig;
-  mrb_value ret;
 
   if (mrb_nil_p(proc))
     return mrb_nil_value();
 
-  rproc = mrb_proc_ptr(proc);
-  orig = rproc->c;
-  rproc->c = mrb->proc_class;
-  ret = mrb_funcall_id(mrb, proc, MRB_SYM(source_location), 0);
-  rproc->c = orig;
-  return ret;
+  return mrb_proc_source_location(mrb, mrb_proc_ptr(proc));
 }
 
 static mrb_value
 method_parameters(mrb_state *mrb, mrb_value self)
 {
   mrb_value proc = mrb_iv_get(mrb, self, MRB_SYM(_proc));
-  struct RProc *rproc;
-  struct RClass *orig;
-  mrb_value ret;
 
   if (mrb_nil_p(proc)) {
     mrb_value rest = mrb_symbol_value(MRB_SYM(rest));
@@ -278,12 +379,7 @@ method_parameters(mrb_state *mrb, mrb_value self)
     return mrb_ary_new_from_values(mrb, 1, &arest);
   }
 
-  rproc = mrb_proc_ptr(proc);
-  orig = rproc->c;
-  rproc->c = mrb->proc_class;
-  ret = mrb_funcall_id(mrb, proc, MRB_SYM(parameters), 0);
-  rproc->c = orig;
-  return ret;
+  return mrb_proc_parameters(mrb, proc);
 }
 
 static mrb_value
@@ -297,6 +393,16 @@ method_to_s(mrb_state *mrb, mrb_value self)
 
   mrb_str_cat_cstr(mrb, str, mrb_obj_classname(mrb, self));
   mrb_str_cat_lit(mrb, str, ": ");
+  if (mrb_type(owner) == MRB_TT_SCLASS) {
+    mrb_value recv = mrb_iv_get(mrb, self, MRB_SYM(_recv));
+    if (!mrb_nil_p(recv)) {
+      mrb_str_concat(mrb, str, recv);
+      mrb_str_cat_lit(mrb, str, ".");
+      mrb_str_concat(mrb, str, name);
+      goto finish;
+    }
+  }
+
   rklass = mrb_class_ptr(klass);
   if (mrb_class_ptr(owner) == rklass) {
     mrb_str_concat(mrb, str, owner);
@@ -304,12 +410,21 @@ method_to_s(mrb_state *mrb, mrb_value self)
     mrb_str_concat(mrb, str, name);
   }
   else {
-    mrb_str_cat_cstr(mrb, str, mrb_class_name(mrb, rklass));
+    rklass = mrb_class_real(rklass); /* skip internal class */
+    mrb_str_concat(mrb, str, mrb_obj_value(rklass));
     mrb_str_cat_lit(mrb, str, "(");
     mrb_str_concat(mrb, str, owner);
     mrb_str_cat_lit(mrb, str, ")#");
     mrb_str_concat(mrb, str, name);
   }
+ finish:;
+  mrb_value loc = method_source_location(mrb, self);
+  if (mrb_array_p(loc) && RARRAY_LEN(loc) == 2) {
+    mrb_str_cat_lit(mrb, str, " ");
+    mrb_str_concat(mrb, str, RARRAY_PTR(loc)[0]);
+    mrb_str_cat_lit(mrb, str, ":");
+    mrb_str_concat(mrb, str, RARRAY_PTR(loc)[1]);
+  }
   mrb_str_cat_lit(mrb, str, ">");
   return str;
 }
@@ -356,7 +471,7 @@ mrb_kernel_method(mrb_state *mrb, mrb_value self)
 
   mrb_search_method_owner(mrb, mrb_class(mrb, self), self, name, &owner, &proc, FALSE);
 
-  me = method_object_alloc(mrb, mrb_class_get(mrb, "Method"));
+  me = method_object_alloc(mrb, mrb_class_get_id(mrb, MRB_SYM(Method)));
   mrb_obj_iv_set(mrb, me, MRB_SYM(_owner), mrb_obj_value(owner));
   mrb_obj_iv_set(mrb, me, MRB_SYM(_recv), self);
   mrb_obj_iv_set(mrb, me, MRB_SYM(_name), mrb_symbol_value(name));
@@ -378,7 +493,7 @@ mrb_module_instance_method(mrb_state *mrb, mrb_value self)
 
   mrb_search_method_owner(mrb, mrb_class_ptr(self), self, name, &owner, &proc, TRUE);
 
-  ume = method_object_alloc(mrb, mrb_class_get(mrb, "UnboundMethod"));
+  ume = method_object_alloc(mrb, mrb_class_get_id(mrb, MRB_SYM(UnboundMethod)));
   mrb_obj_iv_set(mrb, ume, MRB_SYM(_owner), mrb_obj_value(owner));
   mrb_obj_iv_set(mrb, ume, MRB_SYM(_recv), mrb_nil_value());
   mrb_obj_iv_set(mrb, ume, MRB_SYM(_name), mrb_symbol_value(name));
@@ -409,8 +524,8 @@ method_name(mrb_state *mrb, mrb_value self)
 void
 mrb_mruby_method_gem_init(mrb_state* mrb)
 {
-  struct RClass *unbound_method = mrb_define_class(mrb, "UnboundMethod", mrb->object_class);
-  struct RClass *method = mrb_define_class(mrb, "Method", mrb->object_class);
+  struct RClass *unbound_method = mrb_define_class_id(mrb, MRB_SYM(UnboundMethod), mrb->object_class);
+  struct RClass *method = mrb_define_class_id(mrb, MRB_SYM(Method), mrb->object_class);
 
   mrb_undef_class_method(mrb, unbound_method, "new");
   mrb_define_method(mrb, unbound_method, "bind", unbound_method_bind, MRB_ARGS_REQ(1));
