@@ -15,36 +15,24 @@
 #include <mruby/error.h>
 #include <mruby/class.h>
 #include <mruby/throw.h>
+#include <mruby/internal.h>
 #include <mruby/presym.h>
 
-static void
-exc_mesg_set(mrb_state *mrb, struct RException *exc, mrb_value mesg)
+void
+mrb_exc_mesg_set(mrb_state *mrb, struct RException *exc, mrb_value mesg)
 {
-  if (mrb_string_p(mesg)) {
-    exc->flags |= MRB_EXC_MESG_STRING_FLAG;
-    exc->mesg = RSTRING(mesg);
-    mrb_field_write_barrier_value(mrb, (struct RBasic*)exc, mesg);
-  }
-  else {
-    exc->flags &= ~MRB_EXC_MESG_STRING_FLAG;
-    if (mrb_nil_p(mesg)) {
-      exc->mesg = 0;
-    }
-    else {
-      mrb_obj_iv_set(mrb, (struct RObject*)exc, MRB_SYM(mesg), mesg);
-    }
+  if (!mrb_string_p(mesg)) {
+    mesg = mrb_obj_as_string(mrb, mesg);
   }
+  exc->mesg = mrb_obj_ptr(mesg);
+  mrb_field_write_barrier_value(mrb, (struct RBasic*)exc, mesg);
 }
 
-static mrb_value
-exc_mesg_get(mrb_state *mrb, struct RException *exc)
+mrb_value
+mrb_exc_mesg_get(mrb_state *mrb, struct RException *exc)
 {
-  if ((exc->flags & MRB_EXC_MESG_STRING_FLAG) != 0) {
-    return mrb_obj_value(exc->mesg);
-  }
-  else {
-    return mrb_obj_iv_get(mrb, (struct RObject*)exc, MRB_SYM(mesg));
-  }
+  if (exc->mesg == NULL) return mrb_nil_value();
+  return mrb_obj_value(exc->mesg);
 }
 
 MRB_API mrb_value
@@ -52,14 +40,14 @@ mrb_exc_new_str(mrb_state *mrb, struct RClass* c, mrb_value str)
 {
   mrb_ensure_string_type(mrb, str);
 
-  struct RBasic* e = mrb_obj_alloc(mrb, MRB_TT_EXCEPTION, c);
+  struct RException *e = MRB_OBJ_ALLOC(mrb, MRB_TT_EXCEPTION, c);
   mrb_value exc = mrb_obj_value(e);
-  mrb_iv_set(mrb, exc, MRB_SYM(mesg), str);
+  mrb_exc_mesg_set(mrb, e, str);
   return exc;
 }
 
 MRB_API mrb_value
-mrb_exc_new(mrb_state *mrb, struct RClass *c, const char *ptr, size_t len)
+mrb_exc_new(mrb_state *mrb, struct RClass *c, const char *ptr, mrb_int len)
 {
   return mrb_exc_new_str(mrb, c, mrb_str_new(mrb, ptr, len));
 }
@@ -78,7 +66,7 @@ exc_initialize(mrb_state *mrb, mrb_value exc)
   mrb_value mesg;
 
   if (mrb_get_args(mrb, "|o", &mesg) == 1) {
-    exc_mesg_set(mrb, mrb_exc_ptr(exc), mesg);
+    mrb_exc_mesg_set(mrb, mrb_exc_ptr(exc), mesg);
   }
   return exc;
 }
@@ -107,7 +95,7 @@ exc_exception(mrb_state *mrb, mrb_value self)
   if (argc == 0) return self;
   if (mrb_obj_equal(mrb, self, a)) return self;
   exc = mrb_obj_clone(mrb, self);
-  exc_mesg_set(mrb, mrb_exc_ptr(exc), a);
+  mrb_exc_mesg_set(mrb, mrb_exc_ptr(exc), a);
 
   return exc;
 }
@@ -123,7 +111,7 @@ exc_exception(mrb_state *mrb, mrb_value self)
 static mrb_value
 exc_to_s(mrb_state *mrb, mrb_value exc)
 {
-  mrb_value mesg = exc_mesg_get(mrb, mrb_exc_ptr(exc));
+  mrb_value mesg = mrb_exc_mesg_get(mrb, mrb_exc_ptr(exc));
   struct RObject *p;
 
   if (!mrb_string_p(mesg)) {
@@ -136,20 +124,6 @@ exc_to_s(mrb_state *mrb, mrb_value exc)
   return mesg;
 }
 
-/*
- * call-seq:
- *   exception.message   ->  string
- *
- * Returns the result of invoking <code>exception.to_s</code>.
- * Normally this returns the exception's message or name.
- */
-
-static mrb_value
-exc_message(mrb_state *mrb, mrb_value exc)
-{
-  return mrb_funcall_id(mrb, exc, MRB_SYM(to_s), 0);
-}
-
 /*
  * call-seq:
  *   exception.inspect   -> string
@@ -163,10 +137,9 @@ exc_message(mrb_state *mrb, mrb_value exc)
 mrb_value
 mrb_exc_inspect(mrb_state *mrb, mrb_value exc)
 {
-  mrb_value mesg = exc_mesg_get(mrb, mrb_exc_ptr(exc));
   mrb_value cname = mrb_mod_to_s(mrb, mrb_obj_value(mrb_obj_class(mrb, exc)));
-  mesg = mrb_obj_as_string(mrb, mesg);
-  return RSTRING_LEN(mesg) == 0 ? cname : mrb_format(mrb, "%v (%v)", mesg, cname);
+  mrb_value mesg = mrb_exc_mesg_get(mrb, mrb_exc_ptr(exc)); /* string or nil */
+  return (mrb_nil_p(mesg)||RSTRING_LEN(mesg)==0) ? cname : mrb_format(mrb, "%v (%v)", mesg, cname);
 }
 
 void mrb_keep_backtrace(mrb_state *mrb, mrb_value exc);
@@ -187,7 +160,8 @@ set_backtrace(mrb_state *mrb, mrb_value exc, mrb_value backtrace)
       p++;
     }
   }
-  mrb_iv_set(mrb, exc, MRB_SYM(backtrace), backtrace);
+  mrb_exc_ptr(exc)->backtrace = mrb_obj_ptr(backtrace);
+  mrb_field_write_barrier_value(mrb, mrb_basic_ptr(exc), backtrace);
 }
 
 static mrb_value
@@ -221,7 +195,7 @@ static mrb_noreturn void
 exc_throw(mrb_state *mrb, mrb_value exc)
 {
   if (!mrb->jmp) {
-    mrb_p(mrb, exc);
+    mrb_print_error(mrb);
     abort();
   }
   MRB_THROW(mrb->jmp);
@@ -304,7 +278,7 @@ mrb_vformat(mrb_state *mrb, const char *format, va_list ap)
     if (c == '%') {
       if (*p == '!') {
         inspect = TRUE;
-        ++p;
+        p++;
       }
       if (!*p) break;
       switch (*p) {
@@ -534,25 +508,24 @@ exception_call:
   return mesg;
 }
 
-MRB_API void
+MRB_API mrb_noreturn void
 mrb_sys_fail(mrb_state *mrb, const char *mesg)
 {
   struct RClass *sce;
   mrb_int no;
 
   no = (mrb_int)errno;
-  if (mrb_class_defined(mrb, "SystemCallError")) {
-    sce = mrb_class_get(mrb, "SystemCallError");
+  if (mrb_class_defined_id(mrb, MRB_SYM(SystemCallError))) {
+    sce = mrb_class_get_id(mrb, MRB_SYM(SystemCallError));
     if (mesg != NULL) {
-      mrb_funcall(mrb, mrb_obj_value(sce), "_sys_fail", 2, mrb_fixnum_value(no), mrb_str_new_cstr(mrb, mesg));
+      mrb_funcall_id(mrb, mrb_obj_value(sce), MRB_SYM(_sys_fail), 2, mrb_fixnum_value(no), mrb_str_new_cstr(mrb, mesg));
     }
     else {
-      mrb_funcall(mrb, mrb_obj_value(sce), "_sys_fail", 1, mrb_fixnum_value(no));
+      mrb_funcall_id(mrb, mrb_obj_value(sce), MRB_SYM(_sys_fail), 1, mrb_fixnum_value(no));
     }
   }
-  else {
-    mrb_raise(mrb, E_RUNTIME_ERROR, mesg);
-  }
+
+  mrb_raise(mrb, E_RUNTIME_ERROR, mesg);
 }
 
 MRB_API mrb_noreturn void
@@ -603,7 +576,7 @@ mrb_core_init_protect(mrb_state *mrb, void (*body)(mrb_state *, void *), void *o
     err = 0;
   } MRB_CATCH(&c_jmp) {
     if (mrb->exc) {
-      mrb_p(mrb, mrb_obj_value(mrb->exc));
+      mrb_print_error(mrb);
       mrb->exc = NULL;
     }
     else {
@@ -656,6 +629,26 @@ mrb_raise_nomemory(mrb_state *mrb)
   }
 }
 
+MRB_API void
+mrb_print_error(mrb_state *mrb)
+{
+#ifndef MRB_NO_STDIO
+  if (mrb->jmp == NULL) {
+    struct mrb_jmpbuf c_jmp;
+    MRB_TRY(&c_jmp) {
+      mrb->jmp = &c_jmp;
+      mrb_print_backtrace(mrb);
+    } MRB_CATCH(&c_jmp) {
+      /* ignore exception during print_backtrace() */
+    } MRB_END_EXC(&c_jmp);
+    mrb->jmp = NULL;
+  }
+  else {
+    mrb_print_backtrace(mrb);
+  }
+#endif
+}
+
 void
 mrb_init_exception(mrb_state *mrb)
 {
@@ -667,7 +660,6 @@ mrb_init_exception(mrb_state *mrb)
   mrb_define_method(mrb, exception, "exception",       exc_exception,     MRB_ARGS_OPT(1));
   mrb_define_method(mrb, exception, "initialize",      exc_initialize,    MRB_ARGS_OPT(1));
   mrb_define_method(mrb, exception, "to_s",            exc_to_s,          MRB_ARGS_NONE());
-  mrb_define_method(mrb, exception, "message",         exc_message,       MRB_ARGS_NONE());
   mrb_define_method(mrb, exception, "inspect",         mrb_exc_inspect,   MRB_ARGS_NONE());
   mrb_define_method(mrb, exception, "backtrace",       mrb_exc_backtrace, MRB_ARGS_NONE());
   mrb_define_method(mrb, exception, "set_backtrace",   exc_set_backtrace, MRB_ARGS_REQ(1));
