@@ -11,6 +11,7 @@
 #include <mruby/hash.h>
 #include <mruby/string.h>
 #include <mruby/variable.h>
+#include <mruby/internal.h>
 #include <mruby/presym.h>
 
 /*
@@ -69,8 +70,8 @@
 #define AR_MAX_SIZE 16
 #define H_MAX_SIZE EA_MAX_CAPA
 
-mrb_static_assert1(offsetof(struct RHash, iv) == offsetof(struct RObject, iv));
-mrb_static_assert1(AR_MAX_SIZE < (1 << MRB_HASH_AR_EA_CAPA_BIT));
+mrb_static_assert(offsetof(struct RHash, iv) == offsetof(struct RObject, iv));
+mrb_static_assert(AR_MAX_SIZE < (1 << MRB_HASH_AR_EA_CAPA_BIT));
 
 typedef struct hash_entry {
   mrb_value key;
@@ -151,25 +152,25 @@ DEFINE_ACCESSOR(ht, ea_n_used, uint32_t, ea_n_used)
 #else
 DEFINE_FLAG_ACCESSOR(ar, ea_capa, uint32_t, AR_EA_CAPA)
 DEFINE_FLAG_ACCESSOR(ar, ea_n_used, uint32_t, AR_EA_N_USED)
-DEFINE_ACCESSOR(ht, ea_capa, uint32_t, ht->ea_capa)
-DEFINE_ACCESSOR(ht, ea_n_used, uint32_t, ht->ea_n_used)
+DEFINE_ACCESSOR(ht, ea_capa, uint32_t, hsh.ht->ea_capa)
+DEFINE_ACCESSOR(ht, ea_n_used, uint32_t, hsh.ht->ea_n_used)
 #endif
 DEFINE_FLAG_ACCESSOR(ib, bit, uint32_t, IB_BIT)
 DEFINE_ACCESSOR(ar, size, uint32_t, size)
-DEFINE_ACCESSOR(ar, ea, hash_entry*, ea)
+DEFINE_ACCESSOR(ar, ea, hash_entry*, hsh.ea)
 DEFINE_DECREMENTER(ar, size)
 DEFINE_ACCESSOR(ht, size, uint32_t, size)
-DEFINE_ACCESSOR(ht, ea, hash_entry*, ht->ea)
-DEFINE_GETTER(ht, ib, uint32_t*, ht->ib)
+DEFINE_ACCESSOR(ht, ea, hash_entry*, hsh.ht->ea)
+DEFINE_GETTER(ht, ib, uint32_t*, hsh.ht->ib)
 DEFINE_INCREMENTER(ht, size)
 DEFINE_DECREMENTER(ht, size)
 DEFINE_GETTER(h, size, uint32_t, size)
-DEFINE_ACCESSOR(h, ht, hash_table*, ht)
+DEFINE_ACCESSOR(h, ht, hash_table*, hsh.ht)
 DEFINE_SWITCHER(ht, HT)
 
 #define ea_each_used(ea, n_used, entry_var, code) do {                        \
   hash_entry *entry_var = ea, *ea_end__ = entry_var + (n_used);               \
-  for (; entry_var < ea_end__; ++entry_var) {                                 \
+  for (; entry_var < ea_end__; entry_var++) {                                 \
     code;                                                                     \
   }                                                                           \
 } while (0)
@@ -177,7 +178,7 @@ DEFINE_SWITCHER(ht, HT)
 #define ea_each(ea, size, entry_var, code) do {                               \
   hash_entry *entry_var = ea;                                                 \
   uint32_t size__ = size;                                                     \
-  for (; 0 < size__; ++entry_var) {                                           \
+  for (; 0 < size__; entry_var++) {                                           \
     if (entry_deleted_p(entry_var)) continue;                                 \
     --size__;                                                                 \
     code;                                                                     \
@@ -215,40 +216,70 @@ DEFINE_SWITCHER(ht, HT)
 } while (0)
 
 /*
- * `h_check_modified` raises an exception when a dangerous modification is
- * made to `h` by executing `code`.
- *
- * This macro is not called if `h->ht` (`h->ea`) is `NULL` (`Hash` size is
- * zero). And because the `hash_entry` is rather large, `h->ht->ea` and
- * `h->ht->ea_capa` are able to be safely accessed even in AR. This nature
- * is used to eliminate branch of AR or HT.
- *
- * `HT_ASSERT_SAFE_READ` checks if members can be accessed according to its
- * assumptions.
+ * In `h_check_modified()`, in the case of `MRB_NO_BOXING`, `ht_ea()` or
+ * `ht_ea_capa()` for AR may read uninitialized area (#5332). Therefore, do
+ * not use those macros for AR in `MRB_NO_BOXING` (but in the case of
+ * `MRB_64BIT`, `ht_ea_capa()` is the same as `ar_ea_capa()`, so use it).
  */
-#define HT_ASSERT_SAFE_READ(attr_name)                                        \
-  mrb_static_assert1(                                                         \
+#ifdef MRB_NO_BOXING
+# define H_CHECK_MODIFIED_USE_HT_EA_FOR_AR FALSE
+# ifdef MRB_64BIT
+#  define H_CHECK_MODIFIED_USE_HT_EA_CAPA_FOR_AR TRUE
+# else
+#  define H_CHECK_MODIFIED_USE_HT_EA_CAPA_FOR_AR FALSE
+# endif  /* MRB_64BIT */
+#else
+# define H_CHECK_MODIFIED_USE_HT_EA_FOR_AR TRUE
+# define H_CHECK_MODIFIED_USE_HT_EA_CAPA_FOR_AR TRUE
+ /*
+  * `h_check_modified` raises an exception when a dangerous modification is
+  * made to `h` by executing `code`.
+  *
+  * `h_check_modified` macro is not called if `h->hsh.ht` (`h->hsh.ea`) is `NULL`
+  * (`Hash` size is zero). And because the `hash_entry` is rather large,
+  * `h->hsh.ht->ea` and `h->hsh.ht->ea_capa` are able to be safely accessed even for
+  * AR. This nature is used to eliminate branch of AR or HT.
+  *
+  * `HT_ASSERT_SAFE_READ` checks if members can be accessed according to its
+  * assumptions.
+  */
+# define HT_ASSERT_SAFE_READ(attr_name)                                       \
+  mrb_static_assert(                                                          \
     offsetof(hash_table, attr_name) + sizeof(((hash_table*)0)->attr_name) <=  \
     sizeof(hash_entry))
 HT_ASSERT_SAFE_READ(ea);
-#ifdef MRB_32BIT
+# ifdef MRB_32BIT
 HT_ASSERT_SAFE_READ(ea_capa);
-#endif
-#undef HT_ASSERT_SAFE_READ
-#define h_check_modified(mrb, h, code) do {                                   \
-  struct RHash *h__ = h;                                                      \
-  uint32_t mask = MRB_HASH_HT|MRB_HASH_IB_BIT_MASK|MRB_HASH_AR_EA_CAPA_MASK;  \
-  uint32_t flags = h__->flags & mask;                                         \
-  void* tbl__ = (mrb_assert(h__->ht), h__->ht);                               \
-  uint32_t ht_ea_capa__ = ht_ea_capa(h__);                                    \
-  hash_entry *ht_ea__ = ht_ea(h__);                                           \
-  code;                                                                       \
-  if (flags != (h__->flags & mask) ||                                         \
-      tbl__ != h__->ht ||                                                     \
-      ht_ea_capa__ != ht_ea_capa(h__) ||                                      \
-      ht_ea__ != ht_ea(h__)) {                                                \
-    mrb_raise(mrb, E_RUNTIME_ERROR, "hash modified");                         \
-  }                                                                           \
+# endif
+# undef HT_ASSERT_SAFE_READ
+#endif  /* MRB_NO_BOXING */
+
+/*
+ * `h_check_modified` raises an exception when a dangerous modification is
+ * made to `h` by executing `code`.
+ */
+#define h_check_modified(mrb, h, code) do {                                     \
+  struct RHash *h__ = h;                                                        \
+  uint32_t mask__ = MRB_HASH_HT|MRB_HASH_IB_BIT_MASK|MRB_HASH_AR_EA_CAPA_MASK;  \
+  uint32_t flags__ = h__->flags & mask__;                                       \
+  void* tbl__ = (mrb_assert(h__->hsh.ht), h__->hsh.ht);                         \
+  uint32_t ht_ea_capa__ = 0;                                                    \
+  hash_entry *ht_ea__ = NULL;                                                   \
+  if (H_CHECK_MODIFIED_USE_HT_EA_CAPA_FOR_AR || h_ht_p(h__)) {                  \
+    ht_ea_capa__ = ht_ea_capa(h__);                                             \
+  }                                                                             \
+  if (H_CHECK_MODIFIED_USE_HT_EA_FOR_AR || h_ht_p(h__)) {                       \
+    ht_ea__ = ht_ea(h__);                                                       \
+  }                                                                             \
+  code;                                                                         \
+  if (flags__ != (h__->flags & mask__) ||                                       \
+      tbl__ != h__->hsh.ht ||                                                       \
+      ((H_CHECK_MODIFIED_USE_HT_EA_CAPA_FOR_AR || h_ht_p(h__)) &&               \
+       ht_ea_capa__ != ht_ea_capa(h__)) ||                                      \
+      ((H_CHECK_MODIFIED_USE_HT_EA_FOR_AR || h_ht_p(h__)) &&                    \
+       ht_ea__ != ht_ea(h__))) {                                                \
+    mrb_raise(mrb, E_RUNTIME_ERROR, "hash modified");                           \
+  }                                                                             \
 } while (0)
 
 #define U32(v) ((uint32_t)(v))
@@ -263,6 +294,7 @@ static uint32_t ib_bit_to_capa(uint32_t bit);
 static void ht_init(
   mrb_state *mrb, struct RHash *h, uint32_t size,
   hash_entry *ea, uint32_t ea_capa, hash_table *ht, uint32_t ib_bit);
+static void ht_set(mrb_state *mrb, struct RHash *h, mrb_value key, mrb_value val);
 static void ht_set_without_ib_adjustment(
   mrb_state *mrb, struct RHash *h, mrb_value key, mrb_value val);
 
@@ -278,7 +310,7 @@ next_power2(uint32_t v)
   v |= v >> 4;
   v |= v >> 8;
   v |= v >> 16;
-  ++v;
+  v++;
   return v;
 #endif
 }
@@ -296,8 +328,15 @@ obj_hash_code(mrb_state *mrb, mrb_value key, struct RHash *h)
   case MRB_TT_TRUE:
   case MRB_TT_FALSE:
   case MRB_TT_SYMBOL:
+    hash_code = U32(mrb_fixnum(key));
+    break;
   case MRB_TT_INTEGER:
+    if (mrb_fixnum_p(key)) {
+      hash_code = U32(mrb_fixnum(key));
+      break;
+    }
 #ifndef MRB_NO_FLOAT
+    /* fall through */
   case MRB_TT_FLOAT:
 #endif
     hash_code = U32(mrb_obj_id(key));
@@ -387,7 +426,7 @@ ea_compress(hash_entry *ea, uint32_t n_used)
   ea_each_used(ea, n_used, r_entry, {
     if (entry_deleted_p(r_entry)) continue;
     if (r_entry != w_entry) *w_entry = *r_entry;
-    ++w_entry;
+    w_entry++;
   });
 }
 
@@ -493,9 +532,8 @@ ar_set(mrb_state *mrb, struct RHash *h, mrb_value key, mrb_value val)
     if (ea_capa == ea_n_used) {
       if (size == ea_n_used) {
         if (size == AR_MAX_SIZE) {
-          hash_entry *ea = ea_adjust(mrb, ar_ea(h), &ea_capa, EA_MAX_CAPA);
-          ea_set(ea, ea_n_used, key, val);
-          ht_init(mrb, h, ++size, ea, ea_capa, NULL, IB_INIT_BIT);
+          ht_init(mrb, h, size, ar_ea(h), ea_capa, NULL, IB_INIT_BIT);
+          ht_set(mrb, h, key, val);
           return;
         }
         else {
@@ -554,7 +592,7 @@ ar_rehash(mrb_state *mrb, struct RHash *h)
         ea_set(ea, w_size, r_entry->key, r_entry->val);
         entry_delete(r_entry);
       }
-      ++w_size;
+      w_size++;
     }
   });
   mrb_assert(size == w_size);
@@ -718,6 +756,7 @@ ib_bit_for(uint32_t size)
 static uint32_t
 ib_byte_size_for(uint32_t ib_bit)
 {
+  mrb_assert(IB_INIT_BIT <= ib_bit);
   uint32_t ary_size = IB_INIT_BIT == 4 ?
     ib_bit_to_capa(ib_bit) * 2 / IB_TYPE_BIT * ib_bit / 2 :
     ib_bit_to_capa(ib_bit) / IB_TYPE_BIT * ib_bit;
@@ -745,8 +784,9 @@ ht_init(mrb_state *mrb, struct RHash *h, uint32_t size,
 {
   size_t ib_byte_size = ib_byte_size_for(ib_bit);
   size_t ht_byte_size = sizeof(hash_table) + ib_byte_size;
+  ht = (hash_table*)mrb_realloc(mrb, ht, ht_byte_size);
   h_ht_on(h);
-  h_set_ht(h, (hash_table*)mrb_realloc(mrb, ht, ht_byte_size));
+  h_set_ht(h, ht);
   ht_set_size(h, size);
   ht_set_ea(h, ea);
   ht_set_ea_capa(h, ea_capa);
@@ -892,7 +932,13 @@ static void
 ht_rehash(mrb_state *mrb, struct RHash *h)
 {
   /* see comments in `h_rehash` */
-  uint32_t size = ht_size(h), w_size = 0, ea_capa = ht_ea_capa(h);
+  uint32_t size = ht_size(h);
+  if (size <= AR_MAX_SIZE) {
+    ht_to_ar(mrb, h);
+    ar_rehash(mrb, h);
+    return;
+  }
+  uint32_t w_size = 0, ea_capa = ht_ea_capa(h);
   hash_entry *ea = ht_ea(h);
   ht_init(mrb, h, 0, ea, ea_capa, h_ht(h), ib_bit_for(size));
   ht_set_size(h, size);
@@ -933,7 +979,7 @@ h_key_for(mrb_state *mrb, mrb_value key)
 static struct RHash*
 h_alloc(mrb_state *mrb)
 {
-  return (struct RHash*)mrb_obj_alloc(mrb, MRB_TT_HASH, mrb->hash_class);
+  return MRB_OBJ_ALLOC(mrb, MRB_TT_HASH, mrb->hash_class);
 }
 
 static void
@@ -1152,15 +1198,6 @@ mrb_hash_init_copy(mrb_state *mrb, mrb_value self)
   return self;
 }
 
-void
-mrb_hash_check_kdict(mrb_state *mrb, mrb_value self)
-{
-  h_each(mrb_hash_ptr(self), entry, {
-    if (mrb_symbol_p(entry->key)) continue;
-    mrb_raise(mrb, E_ARGUMENT_ERROR, "keyword argument hash with non symbol keys");
-  });
-}
-
 MRB_API mrb_value
 mrb_hash_dup(mrb_state *mrb, mrb_value self)
 {
@@ -1445,6 +1482,7 @@ static mrb_value
 mrb_hash_delete(mrb_state *mrb, mrb_value self)
 {
   mrb_value key = mrb_get_arg1(mrb);
+  mrb->c->ci->mid = 0;
   return mrb_hash_delete_key(mrb, self, key);
 }
 
@@ -1469,7 +1507,7 @@ mrb_hash_shift(mrb_state *mrb, mrb_value hash)
 
   hash_modify(mrb, hash);
   if (h_size(h) == 0) {
-    return hash_default(mrb, hash, mrb_nil_value());
+    return mrb_nil_value();
   }
   else {
     mrb_value del_key, del_val;
@@ -1709,6 +1747,19 @@ mrb_hash_merge(mrb_state *mrb, mrb_value hash1, mrb_value hash2)
   });
 }
 
+static mrb_value
+mrb_hash_merge_m(mrb_state *mrb, mrb_value hash)
+{
+  mrb_int argc;
+  mrb_value *argv;
+
+  mrb_get_args(mrb, "*", &argv, &argc);
+  while (argc--) {
+    mrb_hash_merge(mrb, hash, *argv++);
+  }
+  return hash;
+}
+
 /*
  *  call-seq:
  *    hsh.rehash -> hsh
@@ -1769,4 +1820,5 @@ mrb_init_hash(mrb_state *mrb)
   mrb_define_method(mrb, h, "value?",          mrb_hash_has_value,   MRB_ARGS_REQ(1)); /* 15.2.13.4.27 */
   mrb_define_method(mrb, h, "values",          mrb_hash_values,      MRB_ARGS_NONE()); /* 15.2.13.4.28 */
   mrb_define_method(mrb, h, "rehash",          mrb_hash_rehash,      MRB_ARGS_NONE());
+  mrb_define_method(mrb, h, "__merge",         mrb_hash_merge_m,     MRB_ARGS_REQ(1));
 }
