@@ -13,6 +13,7 @@
 #include <mruby/error.h>
 #include <mruby/data.h>
 #include <mruby/endian.h>
+#include <mruby/internal.h>
 #include <string.h>
 
 #if SIZE_MAX < UINT32_MAX
@@ -162,17 +163,16 @@ read_irep_record_1(mrb_state *mrb, const uint8_t *bin, const uint8_t *end, size_
 #endif
 
       case IREP_TT_BIGINT:
-        pool_data_len = bin_to_uint8(src); /* pool data length */
-        src += sizeof(uint8_t);
+        pool_data_len = bin_to_uint8(src) + 2; /* pool data length */
         if (src + pool_data_len > end) return FALSE;
-        {
+        else {
           char *p;
           pool[i].tt = IREP_TT_BIGINT;
-          p = (char*)mrb_malloc(mrb, pool_data_len+2);
-          memcpy(p, src, pool_data_len+2);
+          p = (char*)mrb_malloc(mrb, pool_data_len);
+          memcpy(p, src, pool_data_len);
           pool[i].u.str = (const char*)p;
         }
-        src += pool_data_len + 2;
+        src += pool_data_len;
         break;
 
       case IREP_TT_FLOAT:
@@ -326,7 +326,7 @@ read_debug_record(mrb_state *mrb, const uint8_t *start, const uint8_t *end, mrb_
   if (bin > end) return MRB_DUMP_GENERAL_FAILURE;
   debug->files = (mrb_irep_debug_info_file**)mrb_calloc(mrb, irep->debug_info->flen, sizeof(mrb_irep_debug_info*));
 
-  for (f_idx = 0; f_idx < debug->flen; ++f_idx) {
+  for (f_idx = 0; f_idx < debug->flen; f_idx++) {
     mrb_irep_debug_info_file *file;
     uint16_t filename_idx;
 
@@ -352,11 +352,12 @@ read_debug_record(mrb_state *mrb, const uint8_t *start, const uint8_t *end, mrb_
         size_t l = sizeof(uint16_t) * (size_t)file->line_entry_count;
 
         if (bin + l > end) return MRB_DUMP_GENERAL_FAILURE;
-        file->lines.ary = (uint16_t *)mrb_malloc(mrb, l);
-        for (l = 0; l < file->line_entry_count; ++l) {
-          file->lines.ary[l] = bin_to_uint16(bin);
+        uint16_t *ary = (uint16_t *)mrb_malloc(mrb, l);
+        for (l = 0; l < file->line_entry_count; l++) {
+          ary[l] = bin_to_uint16(bin);
           bin += sizeof(uint16_t);
         }
+        file->lines.ary = ary;
       } break;
 
       case mrb_debug_line_flat_map: {
@@ -364,21 +365,23 @@ read_debug_record(mrb_state *mrb, const uint8_t *start, const uint8_t *end, mrb_
         size_t n = sizeof(mrb_irep_debug_info_line);
 
         if (bin + c*n > end) return MRB_DUMP_GENERAL_FAILURE;
-        file->lines.flat_map = (mrb_irep_debug_info_line*)mrb_calloc(mrb, c, n);
-        for (size_t l = 0; l < file->line_entry_count; ++l) {
-          file->lines.flat_map[l].start_pos = bin_to_uint32(bin);
+        mrb_irep_debug_info_line *flat_map = (mrb_irep_debug_info_line*)mrb_calloc(mrb, c, n);
+        for (size_t l = 0; l < file->line_entry_count; l++) {
+          flat_map[l].start_pos = bin_to_uint32(bin);
           bin += sizeof(uint32_t);
-          file->lines.flat_map[l].line = bin_to_uint16(bin);
+          flat_map[l].line = bin_to_uint16(bin);
           bin += sizeof(uint16_t);
         }
+        file->lines.flat_map = flat_map;
       } break;
 
       case mrb_debug_line_packed_map: {
         size_t l = (size_t)file->line_entry_count;
 
         if (bin + l > end) return MRB_DUMP_GENERAL_FAILURE;
-        file->lines.packed_map = (uint8_t*)mrb_calloc(mrb, 1, l);
-        memcpy(file->lines.packed_map, bin, file->line_entry_count);
+        uint8_t *packed_map = (uint8_t*)mrb_malloc(mrb, l);
+        memcpy(packed_map, bin, file->line_entry_count);
+        file->lines.packed_map = packed_map;
         bin += file->line_entry_count;
       } break;
 
@@ -432,7 +435,7 @@ read_section_debug(mrb_state *mrb, const uint8_t *start, size_t size, mrb_irep *
   if (bin > end) return MRB_DUMP_GENERAL_FAILURE;
   filenames_obj = mrb_str_new(mrb, NULL, sizeof(mrb_sym) * (size_t)filenames_len);
   filenames = (mrb_sym*)RSTRING_PTR(filenames_obj);
-  for (i = 0; i < filenames_len; ++i) {
+  for (i = 0; i < filenames_len; i++) {
     uint16_t f_len = bin_to_uint16(bin);
     bin += sizeof(uint16_t);
     if (bin + f_len > end) {
@@ -474,7 +477,7 @@ read_lv_record(mrb_state *mrb, const uint8_t *start, mrb_irep *irep, size_t *rec
   if (irep->nlocals == 0) return MRB_DUMP_GENERAL_FAILURE;
   irep->lv = lv = (mrb_sym*)mrb_malloc(mrb, sizeof(mrb_sym) * (irep->nlocals - 1));
 
-  for (i = 0; i + 1 < irep->nlocals; ++i) {
+  for (i = 0; i + 1 < irep->nlocals; i++) {
     uint16_t const sym_idx = bin_to_uint16(bin);
     bin += sizeof(uint16_t);
     if (sym_idx == RITE_LV_NULL_MARK) {
@@ -488,7 +491,7 @@ read_lv_record(mrb_state *mrb, const uint8_t *start, mrb_irep *irep, size_t *rec
     }
   }
 
-  for (i = 0; i < irep->rlen; ++i) {
+  for (i = 0; i < irep->rlen; i++) {
     size_t len;
     int ret;
 
@@ -529,7 +532,7 @@ read_section_lv(mrb_state *mrb, const uint8_t *start, size_t size, mrb_irep *ire
   if (bin > end) return MRB_DUMP_READ_FAULT;
   syms_obj = mrb_str_new(mrb, NULL, sizeof(mrb_sym) * (size_t)syms_len);
   syms = (mrb_sym*)RSTRING_PTR(syms_obj);
-  for (i = 0; i < syms_len; ++i) {
+  for (i = 0; i < syms_len; i++) {
     uint16_t const str_len = bin_to_uint16(bin);
     bin += sizeof(uint16_t);
     if (bin > end) return MRB_DUMP_READ_FAULT;
@@ -666,8 +669,6 @@ irep_error(mrb_state *mrb)
   mrb_exc_set(mrb, mrb_exc_new_lit(mrb, E_SCRIPT_ERROR, "irep load error"));
 }
 
-void mrb_codedump_all(mrb_state*, struct RProc*);
-
 static mrb_value
 load_irep(mrb_state *mrb, struct RProc *proc, mrbc_context *c)
 {
