@@ -10,6 +10,7 @@
 #include <mruby/string.h>
 #include <mruby/dump.h>
 #include <mruby/class.h>
+#include <mruby/internal.h>
 #include <mruby/presym.h>
 
 #ifndef MRB_NO_PRESYM
@@ -110,7 +111,7 @@ sym_inline_unpack(mrb_sym sym, char *buf, mrb_int *lenp)
   for (i=0; i<5; i++) {
     uint32_t bits = sym>>(24-i*6) & 0x3f;
     if (bits == 0) break;
-    buf[i] = pack_table[bits-1];;
+    buf[i] = pack_table[bits-1];
   }
   buf[i] = '\0';
   if (lenp) *lenp = i;
@@ -118,26 +119,6 @@ sym_inline_unpack(mrb_sym sym, char *buf, mrb_int *lenp)
 }
 #endif
 
-static uint8_t
-symhash(const char *key, size_t len)
-{
-    uint32_t hash, i;
-
-    for(hash = i = 0; i < len; ++i) {
-        hash += key[i];
-        hash += (hash << 10);
-        hash ^= (hash >> 6);
-    }
-    hash += (hash << 3);
-    hash ^= (hash >> 11);
-    hash += (hash << 15);
-    return hash & 0xff;
-}
-
-size_t mrb_packed_int_len(uint32_t num);
-size_t mrb_packed_int_encode(uint32_t num, uint8_t *p, uint8_t *pend);
-uint32_t mrb_packed_int_decode(uint8_t *p, uint8_t **newpos);
-
 #define sym_lit_p(mrb, i) (mrb->symflags[i>>3]&(1<<(i&7)))
 #define sym_lit_set(mrb, i) mrb->symflags[i>>3]|=(1<<(i&7))
 #define sym_flags_clear(mrb, i) mrb->symflags[i>>3]&=~(1<<(i&7))
@@ -154,7 +135,7 @@ sym_check(mrb_state *mrb, const char *name, size_t len, mrb_sym i)
   }
   else {
     /* length in BER */
-    symlen = mrb_packed_int_decode((uint8_t*)symname, (uint8_t**)&symname);
+    symlen = mrb_packed_int_decode((const uint8_t*)symname, (const uint8_t**)&symname);
   }
   if (len == symlen && memcmp(symname, name, len) == 0) {
     return TRUE;
@@ -178,7 +159,7 @@ find_symbol(mrb_state *mrb, const char *name, size_t len, uint8_t *hashp)
   i = sym_inline_pack(name, len);
   if (i > 0) return i;
 
-  hash = symhash(name, len);
+  hash = mrb_byte_hash((const uint8_t*)name, len);
   if (hashp) *hashp = hash;
 
   i = mrb->symhash[hash];
@@ -220,7 +201,7 @@ sym_intern(mrb_state *mrb, const char *name, size_t len, mrb_bool lit)
     size_t symcapa = mrb->symcapa;
     if (symcapa == 0) symcapa = 100;
     else symcapa = (size_t)(symcapa * 6 / 5);
-    mrb->symtbl = (const char**)mrb_realloc(mrb, mrb->symtbl, sizeof(char*)*symcapa);
+    mrb->symtbl = (const char**)mrb_realloc(mrb, (void*)mrb->symtbl, sizeof(char*)*symcapa);
     mrb->symflags = (uint8_t*)mrb_realloc(mrb, mrb->symflags, symcapa/8+1);
     memset(mrb->symflags+mrb->symcapa/8+1, 0, (symcapa-mrb->symcapa)/8);
     mrb->symlink = (uint8_t*)mrb_realloc(mrb, mrb->symlink, symcapa);
@@ -232,9 +213,10 @@ sym_intern(mrb_state *mrb, const char *name, size_t len, mrb_bool lit)
     mrb->symtbl[sym] = name;
   }
   else {
-    int ilen = mrb_packed_int_len(len);
+    uint32_t ulen = (uint32_t)len;
+    size_t ilen = mrb_packed_int_len(ulen);
     char *p = (char *)mrb_malloc(mrb, len+ilen+1);
-    mrb_packed_int_encode(len, (uint8_t*)p, (uint8_t*)p+ilen);
+    mrb_packed_int_encode(ulen, (uint8_t*)p, (uint8_t*)p+ilen);
     memcpy(p+ilen, name, len);
     p[ilen+len] = 0;
     mrb->symtbl[sym] = p;
@@ -347,7 +329,7 @@ sym2name_len(mrb_state *mrb, mrb_sym sym, char *buf, mrb_int *lenp)
 
   const char *symname = mrb->symtbl[sym];
   if (!sym_lit_p(mrb, sym)) {
-    size_t len = mrb_packed_int_decode((uint8_t*)symname, (uint8_t**)&symname);
+    uint32_t len = mrb_packed_int_decode((const uint8_t*)symname, (const uint8_t**)&symname);
     if (lenp) *lenp = (mrb_int)len;
   }
   else if (lenp) {
@@ -376,9 +358,9 @@ mrb_free_symtbl(mrb_state *mrb)
       mrb_free(mrb, (char*)mrb->symtbl[i]);
     }
   }
-  mrb_free(mrb, mrb->symtbl);
-  mrb_free(mrb, mrb->symlink);
-  mrb_free(mrb, mrb->symflags);
+  mrb_free(mrb, (void*)mrb->symtbl);
+  mrb_free(mrb, (void*)mrb->symlink);
+  mrb_free(mrb, (void*)mrb->symflags);
 }
 
 void
